{
@abstract(contains the application object TPasdoc)
@author(Erwin Scheuch-Heilig (ScheuchHeilig@t-online.de))
@author(Marco Schmidt (marcoschmidt@geocities.com))
@author(Michael van Canneyt (michael@tfdec1.fys.kuleuven.ac.be))
@created(24 Sep 1999)
@lastmod(20 Apr 2000)
}
unit main;

interface

{$I platform.inc}

{ Wildcard expansion is currently supported for any DOS compiler (FPC/DOS,
  BP) and Delphi.
  Linux doesn't need this, as the shell does the expansion. }

{$ifdef PPC_DELPHI}
  { tell Delphi that this is a console application }
  {$APPTYPE CONSOLE}
  { file search using SysUtils }
  {$define SUPPORT_WILDCARD_EXPANSION}
{$endif}

{$ifdef OS_DOS}
  { file search using DOS unit }
  {$define SUPPORT_WILDCARD_EXPANSION}
  {$define USE_DOS}
{$endif}

{$ifdef PPC_FPC}
  { file search using DOS unit }
  {$define SUPPORT_WILDCARD_EXPANSION}
  {$define USE_DOS}
{$endif}

uses
  {$ifdef USE_DOS}
  DOS,
  {$endif}
  {$ifdef PPC_DELPHI}
  SysUtils,
  {$endif}
  appinfo, arrays, chars, filestre, gendoc, html, items, msg, numbers,
  objects, parsing, streams, tex, texts, time;

type
  { all supported output formats }
  TOutputFormat = (OUTPUT_NONE, OUTPUT_LATEX, OUTPUT_HTML);
  { pointer to @link(TPasDoc) }
  PPasDoc = ^TPasDoc;
  { The main object in the pasdoc application; first scans parameters, then
    parses files.
    All parsed units are then given to documentation generator,
    which creates one or more documentation output files. }
  TPasDoc = object(TObject)
    { all description file names }
    DescriptionFiles: PStringCollection;
    { directory where documentation will be written to }
    DestDir: String;
    { all file names which have been collected from the program
      parameters }
    FileNames: PStringCollection;
    { collection of @link(PString) values of all directories where pasdoc is
      supposed to search for include files }
    IncludeFilePaths: PCollection;
    { defines if the private parts (fields and methods) of objects are to be
      included in the documentation (default is false) }
    IncludePrivate: Boolean;
    KeepLineFeeds: Boolean;
    { language used to write documentation }
    Language: TLanguage;
    { stores if the Language field was ever initialized from the parameters }
    LanguageInitialized: Boolean;
    { by default, all HTML documents contain a link on the bottom stating
      pasdoc version, time this page was generated and a link to the pasdoc
      homepage; set NoGeneratorInfo to true to prevent this }
    NoGeneratorInfo: Boolean;
    { if true: leave out statements at the beginning and end of TeX document
      to enable users to simply \include the pasdoc output; default is false }
    NoTexHeadersAndFooters: Boolean;
    { output format to be used (default is HTML) }
    OutputFormat: TOutputFormat;
    { list of standard compiler directives }
    StdDirectives: PStringCollection;
    { all PUnit objects which have been created from the list of file names }
    Units: PItemCollection;
    { Creates object and sets fields to default values. }
    constructor Init;
    { }
    destructor Done; virtual;
    { Adds S to the list of description files that will be parsed.
      Will compare S to strings in list and will not include it if already
      present.
      However, this will not avoid duplicates on file systems where case is
      not regarded (VFAT etc.). }
    function AddDescriptionFile(S: String): Boolean;
    { Adds parameter S to the internal list of standard directives, which
      are considered present whenever parsing of a unit starts. }
    function AddStandardDirective(S: String): Boolean;
    { Creates a @link(TUnit) object from the stream and adds it to
	  @link(Units). }
    procedure HandleStream(InputStream: PInputStream);
    { Scans the program parameters, creating a list of file names and
      regarding switches. }
    procedure HandleParams;
    { Loads names of Pascal unit source code files from a text file.
      Adds all file names to @link(FileNames). }
    procedure LoadFileNames(FileName: String);
    { Loads a list of standard directives (one per row) from text file
      FileName.
      These directives are added to the standard directives list
      @link(StdDirectives). }
    procedure LoadStandardDirectives(FileName: String);
    { Calls @link(HandleStream) for each file name in @link(FileNames). }
    procedure ParseFiles;
    { Prints a detailed list of switches to standard output. }
    procedure PrintHelp;
    { Prints a usage clause to standard output. }
    procedure PrintUsage;
    { Searches the description of each PUnit item in the collection for an
      excluded tag.
      If one is found, the item is removed from the collection.
      If not, the fields, methods and properties collections are called
      with RemoveExcludedItems
      If the collection is empty after removal of all items, it is disposed
      of and the variable is set to nil. }
    procedure RemoveExcludedItems(var C: PItemCollection);
    { Does all the work. }
    procedure Run;
    { Searches for descr tags in the comments of all TItem objects in C. }
    procedure SearchDescrFileTags(var C: PItemCollection);
    { Sets language to NewLanguage. }
    procedure SetLanguage(const NewLanguage: TLanguage);
  end;

implementation

const
  { the default output language (type @link(TLanguage) }
  DEFAULT_LANGUAGE: TLanguage = LANG_ENGLISH;
  { the output format to be used when none is specified on command line }
  DEFAULT_OUTPUT_FORMAT: TOutputFormat = OUTPUT_HTML;
  { names of all output formats }
  OUTPUT_FORMAT_NAMES: Array[TOutputFormat] of String =
    ('', 'LaTeX', 'HTML');
  { the string used to separate paths, a backslash on Windows and a forward
    slash on Unix systems }
  {$ifdef OS_LINUX}
  PATH_SEPARATOR = '/';
  {$else}
  PATH_SEPARATOR = '\';
  {$endif}

{ find the right-most character 'CharToFind' within string 'Str'
  returns 0 if character not found at all }
function RScan(CharToFind: char; Str: String):integer;
var
  CurrPos: integer;
begin
  CurrPos := Length(Str);
  while (CurrPos > 0) and (Str[CurrPos] <> CharToFind) do
    dec(CurrPos);
  RScan := currPos;
end;

constructor TPasDoc.Init;
begin
  inherited Init;
  { check for Unicode }
  if (SizeOf(Char) <> 1)
  then
    begin
      PrintLn(1, 'Error: Size of Char is larger than one byte, cannot handle ' +
        'this case.');
      Halt(1);
    end;
  DestDir := '';
  DescriptionFiles := nil;
  FileNames := nil;
  IncludeFilePaths := nil;
  IncludePrivate := false;
  KeepLineFeeds := false;
  Language := LANG_UNDEFINED;
  LanguageInitialized := false;
  OutputFormat := DEFAULT_OUTPUT_FORMAT;
  NoGeneratorInfo := false;
  NoTexHeadersAndFooters := false;
  StdDirectives := nil;
  Units := nil;
end;

destructor TPasDoc.Done;
begin
  if Assigned(DescriptionFiles)
  then Dispose(DescriptionFiles, Done);
  if Assigned(FileNames)
  then Dispose(FileNames, Done);
  if Assigned(Units)
  then Dispose(Units, Done);
  inherited Done;
end;

function TPasDoc.AddDescriptionFile(S: String): Boolean;
var
  PS: PString;
begin
  AddDescriptionFile := false;
  if (not Assigned(DescriptionFiles)) then
  begin
    DescriptionFiles := New(PStringCollection, Init(16, 16));
    if (not Assigned(DescriptionFiles)) then
    begin
      PrintLn(2, 'Errorr: not enough memory to create description file list');
      exit;
    end;
  end;
  if (DescriptionFiles^.IndexOf(@S) <> -1) then
  begin
    AddDescriptionFile := true;
    exit;
  end;
  PS := NewStr(S);
  if Assigned(PS)
  then DescriptionFiles^.Insert(PS)
  else PrintLn(2, 'Error - could not create description file name string');
  AddDescriptionFile := Assigned(PS);
end;

function TPasDoc.AddStandardDirective(S: String): Boolean;
var
  PS: PString;
begin
  { convert S to upper case for convenient comparison }
  StringToUpper(S, S);
  { check if S is already included - if so, leave }
  if (Assigned(StdDirectives) and (StdDirectives^.IndexOf(@S) <> -1))
  then
    begin
      AddStandardDirective := true;
      exit;
    end;
  { create StdDirectives collection if necessary }
  if (not Assigned(StdDirectives))
  then
    begin
      StdDirectives := New(PStringCollection, Init(16, 16));
      if (not Assigned(StdDirectives))
      then
        begin
          PrintLn(1, 'Error: Could not create list of standard directives.');
          Halt(1);
        end;
    end;
  { create PString version of S and insert it into collection }
  PS := NewStr(S);
  if (not Assigned(PS))
  then
    begin
      PrintLn(1, 'Error: Could not create new string for directive list');
      Halt(1);
    end;
  StdDirectives^.Insert(PS);
  AddStandardDirective := true;
end;

procedure TPasDoc.HandleParams;
var
  ErrorCode: Integer;
  I: LongInt;
  J: LongInt;
  N: LongInt;
  P: PString;
  Path: String;
  S: String;
  VerbosityLevel: Integer;
  {$ifdef USE_DOS}
  SR: DOS.SearchRec;
  {$endif}
  {$ifdef PPC_DELPHI}
  SearchResult: Integer;
  SR: SysUtils.TSearchRec;
  {$endif}
begin
  N := ParamCount;
  if (N < 1)
  then
    begin
      PrintUsage;
      PrintLn(1, 'Try ''' + ParamStr(0) + ' --help'' for more information.');
      Halt(0);
    end;
  FileNames := New(PStringCollection, Init(16, 16));
  OutputFormat := OUTPUT_NONE;
  VerbosityLevel := DEFAULT_VERBOSITY_LEVEL;
  I := 1;
  repeat
    S := ParamStr(I);
    if (S[1] = '-') then
    begin
      System.Delete(S, 1, 1);
      if (S = '-help') or (S = 'H') then
      begin
        PrintUsage;
        PrintHelp;
        Halt(0);
      end
      else
      if (S = '-version') or (S = 'V')
      then
        begin
          PrintLn(1, APP_NAME_AND_VERSION);
          Halt(0);
        end
      else
      if (S = 'a') or (S = '-catalan')
      then SetLanguage(LANG_CATALAN)
      else
      if (S = 'b') or (S = '-brasilian')
      then SetLanguage(LANG_BRASILIAN)
      else
      if (S = 'c') or (S = '-descrfile') then
      begin
        if (I = N) then
        begin
          PrintLn(1, 'Error: No file name specified after -c switch.');
          Halt(1);
        end;
        Inc(I);
        if (not AddDescriptionFile(ParamStr(I))) then
        begin
          PrintLn(1, 'Error: Could not add description file.');
          Halt(1);
        end;
      end
      else
      if (S = 'd') or (S = '-directive')
      then
        begin
          if (I = N)
          then
            begin
              PrintLn(1, 'Error: No directive specified after -d switch.');
              Halt(1);
            end;
          Inc(I);
          if (not AddStandardDirective(ParamStr(I)))
          then
            begin
              PrintLn(1, 'Error: Could not add standard directive.');
              Halt(1);
            end;
        end
      else
      if (S = 'e') or (S = '-english')
      then SetLanguage(LANG_ENGLISH)
      else
      if (S = 'f') or (S = '-directivefile')
      then
        begin
          if (I = N)
          then
            begin
              PrintLn(1, 'Error: No directive file specified after -f switch.');
              Halt(1);
            end;
          Inc(I);
          LoadStandardDirectives(ParamStr(I));
        end
      else
      if (S = 'g') or (S = '-german')
      then SetLanguage(LANG_GERMAN)
      else
      if (S = 'h') or (S = '-html')
      then
        begin
          if (OutputFormat <> OUTPUT_NONE) and (OutputFormat <> OUTPUT_HTML)
          then
            begin
              PrintLn(1, 'Error: Cannot handle more than one documentation ' +
                'output type.');
              Halt(1);
            end;
          OutputFormat := OUTPUT_HTML;
        end
      else
      if (S = 'i') or (S = '-spanish')
      then SetLanguage(LANG_SPANISH)
      else
      if (S = 'j') or (S = '-notexheaders')
      then NoTexHeadersAndFooters := true
      else
      if (S = 'k') or (S = '-keeplinefeeds')
      then KeepLineFeeds := true
      else
      if (S = 'l') or (S = '-latex')
      then
        begin
          if (OutputFormat <> OUTPUT_NONE) and (OutputFormat <> OUTPUT_LATEX)
          then
            begin
              PrintLn(1, 'Error: Cannot handle more than one documentation output type.');
              Halt(1);
            end;
          OutputFormat := OUTPUT_LATEX;
        end
      else
      if (S = 'm') or (S = '-dutch')
      then SetLanguage(LANG_DUTCH)
      else
      if (S = 'n') or (S = '-nohomepage')
      then NoGeneratorInfo := true
      else
      if (S = 'o') or (S = '-outputdirectory')
      then
        begin
          if (I = N)
          then
            begin
              PrintLn(1, 'Error: No directory specified after -o switch.');
              Halt(1);
            end;
          Inc(I);
          DestDir := ParamStr(I);
          if (Length(DestDir) > 0) and
             (DestDir[Length(DestDir)] <> PATH_SEPARATOR)
          then DestDir := DestDir + PATH_SEPARATOR;
        end
      else
      if (S = 'p') or (S = '-includeprivate')
      then IncludePrivate := true
      else
      if (S = 'r') or (S = '-french')
      then SetLanguage(LANG_FRENCH)
      else
      if (S = 's') or (S = '-sourcefilenames')
      then
        begin
          if (I = N)
          then
            begin
              PrintLn(1, 'Error: No file name specified after -s switch.');
              Halt(1);
            end;
          Inc(I);
          LoadFileNames(ParamStr(I));
        end
      else
      if (S = 'u') or (S = '-includefilepath') then
      begin
        if (I = N) then
        begin
          PrintLn(1, 'Error: No directory specified after -u switch.');
          Halt(1);
        end;
        Inc(I);
        S := ParamStr(I);
        if (Length(S) > 0) and (S[Length(S)] <> PATH_SEPARATOR)
        then S := S + PATH_SEPARATOR;
        if (not Assigned(IncludeFilePaths)) then
        begin
          IncludeFilePaths := New(PCollection, Init(16, 16));
          if (not Assigned(IncludeFilePaths)) then
          begin
            PrintLn(1, 'Error: Could not create list of include file paths.');
            Halt(1);
          end;
        end;
        P := NewStr(S);
        if (not Assigned(P)) then
        begin
          PrintLn(1, 'Error: Could not create string for include file path.');
          Halt(1);
        end;
        IncludeFilePaths^.Insert(P);
      end
      else
      if (S = 'v') or (S = '-verbose')
      then
        begin
          if (I = N)
          then
            begin
              PrintLn(1, 'Error: No verbosity level (integer >= 0) specified after -v switch.');
              Halt(1);
            end;
          Inc(I);
          Val(ParamStr(I), VerbosityLevel, ErrorCode);
          if (ErrorCode <> 0) or (VerbosityLevel < 0) then
          begin
            PrintLn(1, 'Error - verbosity level value must be an integer >= 0');
            Halt(1);
          end;
        end
      else
        begin
          PrintLn(1, 'Error: Unknown switch ("-' + S + '")');
          Halt(1);
        end;
    end
    else
    begin
      { it's not a switch, so it must be a unit source code file name
	  or a wildcard }
      { check for wildcards }
      if (System.Pos('*', S) > 0) or (System.Pos('?', S) > 0) then
      begin
        {$ifdef SUPPORT_WILDCARD_EXPANSION}
        Path := '';
		{ Path = everything before right-most path delimiter char }
		if (RScan(PATH_SEPARATOR, S) > 0)
        then Path := System.Copy(S, 1, RScan(PATH_SEPARATOR, S));
        {$ifdef USE_DOS}
        DOS.FindFirst(S, 63, SR);
        while (DOS.DosError = 0) do
        begin
          if ((SR.Attr and 24) = 0)
          then FileNames^.Insert(NewStr(Path + SR.Name));
          DOS.FindNext(SR);
        end;
        {$endif}
        {$ifdef PPC_DELPHI}
        SearchResult := SysUtils.FindFirst(S, 63, SR);
        while (SearchResult = 0) do
        begin
          if ((SR.Attr and 24) = 0)
          then FileNames^.Insert(NewStr(Path + SR.Name));
          SearchResult := FindNext(SR);
        end;
        FindClose(SR);
        {$endif}
        {$else}
        PrintLn(1, 'Warning: Wildcard expansion not supported, skipping ',
		  S);
        {$endif}
      end
      else
      begin
        P := NewStr(S);
        J := FileNames^.IndexOf(P);
        if (J < 0)
        then FileNames^.Insert(P)
        else DisposeStr(P);
      end;
    end;
    Inc(I);
  until (I > N);
  SetVerbosityLevel(VerbosityLevel);
  if (FileNames^.Count < 1)
  then
  begin
    PrintLn(1, 'Error: No file has been specified.');
    Halt(1);
  end;
  PrintLn(2, APP_FIRST_LINE);
  { if no output format has been defined on the command line, pick
    default output format, show warning }
  if (OutputFormat = OUTPUT_NONE) then
  begin
    PrintLn(2, 'Warning: No output format specified, using default ("' +
      OUTPUT_FORMAT_NAMES[DEFAULT_OUTPUT_FORMAT] + '").');
    OutputFormat := DEFAULT_OUTPUT_FORMAT;
  end;
  if (Language = LANG_UNDEFINED) then
  begin
    SetLanguage(DEFAULT_LANGUAGE);
    PrintLn(2, 'Warning: No output language specified, using default ("' +
    	LANGUAGE_NAMES[Language] + '").');
  end;
{  if Assigned(IncludeFilePaths) and (IncludeFilePaths^.Count > 0) then
  begin
    for I := 0 to IncludeFilePaths^.Count - 1 do
    begin
      P := IncludeFilePaths^.At(I);
      if Assigned(P)
      then WriteLn(I, ' ', P^);
    end;
  end;}
end;

procedure TPasDoc.HandleStream(InputStream: PInputStream);
var
  P: PParser;
  U: PUnit;
begin
  P := New(PParser, Init(InputStream, StdDirectives, IncludeFilePaths));
  if (not Assigned(P))
  then
    begin
      PrintLn(1, InputStream^.GetName + ' Error: Could not create parser.');
      exit;
    end;
  if P^.ParseUnit(U)
  then
    begin
      if (not IncludePrivate)
      then U^.RemovePrivateItems;
      if (not Assigned(Units))
      then
      begin
        Units := New(PItemCollection, Init(16, 16));
      end;
      Units^.Insert(U);
    end
  else
    begin
      PrintLn(1, 'Error parsing: ' + P^.ErrorMessage);
    end;
  Dispose(P, Done);
end;

procedure TPasDoc.LoadFileNames(FileName: String);
var
  NF: PFileInputStream;
  L: String;
  S: PString;
begin
  NF := New(PFileInputStream, Init(FileName));
  if (not Assigned(NF)) then
  begin
    PrintLn(2, 'Error: Could not open file "' + FileName +
        '" to read file names.');
    Halt(1);
  end;
  while (NF^.HasData) do
  begin
    NF^.ReadLine(L);
    S := NewStr(L);
    if (not Assigned(S)) then
    begin
      PrintLn(2, 'Error: Not enough memory to create list of file names.');
      Halt(1);
    end;
    if (FileNames^.IndexOf(@S) = -1)
    then FileNames^.Insert(S)
    else DisposeStr(S);
  end;
  Dispose(NF, Done);
end;

procedure TPasDoc.LoadStandardDirectives(FileName: String);
var
  DF: PFileInputStream;
  L: String;
begin
  DF := New(PFileInputStream, Init(FileName));
  if (not Assigned(DF))
  then
    begin
      PrintLn(1, 'Error: Could not open file "' + FileName +
        '" to read standard directives.');
      Halt(1);
    end;
  while (DF^.HasData) do
  begin
    DF^.ReadLine(L);
    if (not AddStandardDirective(L))
    then Break;
  end;
  Dispose(DF, Done);
end;

procedure TPasDoc.ParseFiles;
var
  I: LongInt;
  P: PString;
  InputStream: PFileInputStream;
begin
  if (not Assigned(FileNames)) or (FileNames^.Count < 1)
  then exit;
  I := 0;
  repeat
    P := FileNames^.At(I);
    PrintLn(3, 'Loading "' + P^ + '"...');
    InputStream := New(PFileInputStream, Init(P^));
    if (not Assigned(InputStream)) then
    begin
      PrintLn(2, 'Warning: Could not open "' + P^ + '" for reading, skipping.');
    end
    else HandleStream(InputStream);
    Inc(I);
  until (I = FileNames^.Count);
end;

procedure TPasDoc.PrintHelp;
begin
  PrintLn(1, '');
  PrintLn(1, 'General switches');
  PrintLn(1, ' -H, --help                   Show this screen and terminate');
  PrintLn(1, ' -V, --version                Show program version and terminate');
  PrintLn(1, ' -v, --verbosity LEVEL        Set verbosity to LEVEL (default 2)');
  PrintLn(1, 'Parser switches');
  PrintLn(1, ' -c, --descrfile FILE         Read descriptions from FILE');
  PrintLn(1, ' -d, --directive DIRECTIVE    Add a directive');
  PrintLn(1, ' -f, --directivefile FILE     Add directives from FILE');
  PrintLn(1, ' -p, --includeprivate         Include private fields/methods/properties');
  PrintLn(1, ' -s, --sourcefilenames FILE   Read file names from FILE');
  PrintLn(1, ' -u, --includefilepath DIR    Search for include files in DIR');
  PrintLn(1, 'Documentation switches');
  PrintLn(1, ' Formats');
  PrintLn(1, '  -h, --html                  HTML output (default)');
  PrintLn(1, '  -l, --latex                 LaTeX output');
  PrintLn(1, ' Languages');
  PrintLn(1, '  -b, --brasilian             Brasilian Portuguese');
  PrintLn(1, '  -a, --catalan               Catalan');
  PrintLn(1, '  -m, --dutch                 Dutch');
  PrintLn(1, '  -e, --english               English (default)');
  PrintLn(1, '  -r, --french                French');
  PrintLn(1, '  -g, --german                German');
  PrintLn(1, '  -i, --spanish               Spanish');
  PrintLn(1, ' Other');
  PrintLn(1, '  -j, --notexheaders          No headers and footers in LaTeX output');
  PrintLn(1, '  -k, --keeplinefeeds         Keep description line feeds as found in source');
  PrintLn(1, '  -n, --nohomepage            No pasdoc homepage link in HTML output');
  PrintLn(1, '  -o, --outputdirectory DIR   Write doc file(s) to directory DIR');
  PrintLn(1, '');
  PrintLn(1, 'Visit the pasdoc homepage: ' + APP_HOMEPAGE);
end;

procedure TPasDoc.PrintUsage;
begin
  PrintLn(1, APP_FIRST_LINE);
  PrintLn(1, 'Usage: ' + ParamStr(0) + ' {<option(s)>} <FILEs>');
  PrintLn(1, 'Documentation generator for Pascal unit source code files.');
end;

procedure TPasDoc.RemoveExcludedItems(var C: PItemCollection);
var
  I: LongInt;
  Offs1: LongInt;
  Offs2: LongInt;
  Offs3: LongInt;
  P: PItem;
begin
  if (not Assigned(C))
  then exit;
  I := 0;
  while (I < C^.Count) do
  begin
    P := C^.At(I);
    Offs1 := 0;
    if Assigned(P) and Assigned(P^.Description) and
      (P^.Description^.FindTag('EXCLUDE', Offs1, Offs2, Offs3)) then
    begin
      PrintLn(3, 'Hint: Excluding item ' + P^.Name);
      C^.AtFree(I);
    end
    else
      begin
        { P has no excluded tag; but if it is a class, interface, object or
          unit, one of its parts may be excluded }
        if (TypeOf(P^) = TypeOf(TCIO))
        then
          begin
{            RemoveExcludedItems(PCIO(P)^.Fields);
            RemoveExcludedItems(PCIO(P)^.Properties);
            RemoveExcludedItems(PCIO(P)^.Methods);}
          end
        else
          begin
            if (TypeOf(P^) = TypeOf(TUnit))
            then
              begin
{                RemoveExcludedItems(PUnit(P)^.CIO);
                RemoveExcludedItems(PUnit(P)^.Constants);
                RemoveExcludedItems(PUnit(P)^.FuncsProcs);
                RemoveExcludedItems(PUnit(P)^.Types);
                RemoveExcludedItems(PUnit(P)^.Variables);}
              end;
          end;
        Inc(I);
      end;
  end;
  if (C^.Count = 0) then
  begin
    Dispose(C, Done);
    C := nil;
  end;
end;

procedure TPasDoc.Run;
var
  Generator: PDocGenerator;
  T1: TTime;
  T2: TTime;
begin
  T1.Init;
  HandleParams;
  ParseFiles;
  {P^.SearchDescrFileTags(P^.Units);}
  RemoveExcludedItems(Units);
  { check if parsing was successful }
  if (not Assigned(Units)) or (Units^.Count < 1) then
  begin
    PrintLn(1, 'Error: At least one unit must have been successfully ' +
      'parsed to write docs.');
    Halt(1);
  end;
  PrintLn(3, 'Creating ' + OUTPUT_FORMAT_NAMES[OutputFormat] +
    ' documentation file(s)...');
  { create desired output generator }
  case OutputFormat of
    OUTPUT_HTML:
      Generator := New(PHTMLDocGenerator, Init);
    OUTPUT_LATEX:
    begin
      Generator := New(PTexDocGenerator, Init);
      PTexDocGenerator(Generator)^.NoTexHeadersAndFooters :=
        NoTexHeadersAndFooters;
    end;
    else
      begin
        Generator := nil; { to prevent 'uninitialized' warnings }
        PrintLn(1, 'Error: Output format unknown.');
        Halt(1);
      end;
  end;
  { copy some fields from P to Generator }
  Generator^.Units := Units;
  Generator^.KeepLineFeeds := KeepLineFeeds;
  Generator^.Language := Language;
  Generator^.DestDir := DestDir;
  Generator^.NoGeneratorInfo := NoGeneratorInfo;
  Generator^.Units := Units;
  Generator^.BuildLinks;
  Generator^.ExpandDescriptions;
  Generator^.LoadDescriptionFiles(DescriptionFiles);
  Generator^.WriteDocumentation;
  Dispose(Generator, Done);
  T2.Init;
  PrintLn(2, 'Done, worked ' + IntToString(T2.GetDiffSeconds(T1)) +
    ' second(s).');
end;

procedure TPasDoc.SearchDescrFileTags(var C: PItemCollection);
var
  Found: Boolean;
  I: LongInt;
  Offs1: LongInt;
  Offs2: LongInt;
  Offs3: LongInt;
  P: PItem;
  S: String;
begin
  if (not Assigned(C))
  then exit;
  I := 0;
  while (I < C^.Count) do
  begin
    P := C^.At(I);
    Inc(I);
    if (not Assigned(P))
    then Continue;
    if Assigned(P^.Description)
    then
      begin
        Offs1 := 0;
        repeat
          Found := P^.Description^.FindTag('DESCRFILE', Offs1, Offs2, Offs3);
          if Found
          then
            begin
              P^.Description^.ExtractTag(Offs1, Offs2, Offs3, S);
              PrintLn(3, 'Adding description file "' + S + '"');
              AddDescriptionFile(S);
              Offs1 := Offs3 + 1;
            end;
        until (not Found);
      end;
    if (TypeOf(P^) = TypeOf(TCIO))
    then
      begin
        SearchDescrFileTags(PCIO(P)^.Fields);
{        SearchDescrFileTags(PCIO(P)^.Methods);
        SearchDescrFileTags(PCIO(P)^.Properties);}
      end;
    if (TypeOf(P^) = TypeOf(TUnit))
    then
      begin
{        SearchDescrFileTags(PUnit(P)^.CIO);
        SearchDescrFileTags(PUnit(P)^.Constants);
        SearchDescrFileTags(PUnit(P)^.FuncsProcs);
        SearchDescrFileTags(PUnit(P)^.Types);
        SearchDescrFileTags(PUnit(P)^.Variables);}
      end;
  end;
end;

procedure TPasDoc.SetLanguage(const NewLanguage: TLanguage);
begin
  if LanguageInitialized and (Language <> NewLanguage) then
  begin
    PrintLn(1, 'Error - You can only define one language.');
    Halt(1);
  end;
  Language := NewLanguage;
  LanguageInitialized := true;
end;

begin
end.
